// Copyright 2015 The go4 Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// +build plan9

package osutil

import (
	"fmt"
	"os"
	"path/filepath"
	"syscall"
)

func executable() (string, error) {
	fn := fmt.Sprintf("/proc/%d/text", os.Getpid())
	f, err := os.Open(fn)
	if err != nil {
		return "", err
	}
	defer f.Close()
	p, err := syscall.Fd2path(int(f.Fd()))
	return filepath.Clean(p), err
}
